-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

-- Modified by Gamergull (Thomas W.)

-- [[ STORE FREQUENTLY USED FUNCTIONS IN UPVALUES ]] --
local max = math.max
local min = math.min
local sin = math.sin
local asin = math.asin
local pi = math.pi
local abs = math.abs
local sqrt = math.sqrt
local floor = math.floor
local tableInsert = table.insert
local tableRemove = table.remove
local strFormat = string.format
---------------------------------
local scriptai = nil

local M = {}

M.mode = 'disabled' -- this is the main mode
M.manualTargetName = nil
M.debugMode = 'off'
M.speedMode = nil
M.routeSpeed = nil
M.extAggression = 0.3
M.cutOffDrivability = 0

-- [[ Simulation time step]] --
local dt

-- [[ ENVIRONMENT VARIABLES ]] --
local g = abs(obj:getGravity())
local gravityDir = vec3(0, 0, -1)
local gravityVec = gravityDir * g
----------------------------------

-- [[ PERFORMANCE RELATED ]] --
local aggression = 1
local aggressionMode
local forceLateral = 0
local distPastEdge = 0
--------------------------------------------------------------------------

-- [[ AI DATA: POSITION, CONTROL, STATE ]] --
local aiPos = vec3(obj:getFrontPosition())
--aiPos.z = BeamEngine:getSurfaceHeightBelow(aiPos)
local aiDirVec = vec3(obj:getDirectionVector())
local prevDirVec = vec3(aiDirVec)
local aiUpVec = vec3(obj:getDirectionVectorUp())
local aiRightVec = vec3(0, 0, 0)
local aiVel = vec3(obj:getVelocity())
local aiSpeed = aiVel:length()
local targetSpeedSmoother = newTemporalSmoothingNonLinear(math.huge, 0.2)
local aiWidth -- this comes back zero at this point
local aiLength -- this comes back zero at this point
local aiDeviation = 0
local aiDeviationSmoother = newTemporalSmoothing(1)
local aiCannotMoveTime = 0
local aiForceGoFrontTime = 0
local staticFrictionCoef = 1

local forces = {}
local threewayturn = {state = 0, speedDifInt = 0}
local lastCommand = {steering = 0, throttle = 0, brake = 0, parkingbrake = 0}

local driveInLaneFlag = false
local internalState = 'onroad'

local validateInput = nop
------------------------------

-- [[ CRASH DETECTION ]] --
local crashTime = 0
local crashManeuver = 0
local crashDir = nil

-- [[ OPPONENT DATA ]] --
local player
local plPrevVel
local chaseData = {suspectState = nil, suspectStoppedTimer = 0, distanceToSuspect = nil, playerSeg = nil, playerXnormOnSeg = nil, playerRoad = nil}

-- [[ SETTINGS, PARAMETERS, AUXILIARY DATA ]] --
local map
local currentRoute
local minPlanCount = 3
local targetWPName

local wpList, manualPath, speedList
local race, noOfLaps

local targetObjectSelectionMode

local edgeDict

------------------------------

-- [[ TRAFFIC ]] --
local currPathIdx = 0
local damageFlag = false
local hornFlag = false
local trafficTable = {}
local trafficBlock = {timer = 0, coef = 0, limit = 6, horn = 0}
local trafficSide = {timer = 0, cTimer = 0, side = 1, timerRange = 6}
local trafficSignal = {hTimer = 0, hLimit = 1}
local intersection = {timer = 0, turn = 0}
local avoidCars = 'on'
M.avoidCarsMaster = 'auto'
local changePlanTimer = 0

-----------------------

-- [[ HEAVY DEBUG MODE ]] --
local speedRecordings = {}
local trajecRec = {last = 0}
local routeRec = {last = 0}
local labelRenderDistance = 10
------------------------------

local function aistatus(status, category)
  guihooks.trigger("AIStatusChange", {status=status, category=category})
end

local function getState()
  return M
end

local function stateChanged()
  if playerInfo.anyPlayerSeated then
    guihooks.trigger("AIStateChange", getState())
  end
end

local function setSpeed(speed)
  if type(speed) ~= 'number' then M.routeSpeed = nil else M.routeSpeed = speed end
end

local function setSpeedMode(speedMode)
  if speedMode == 'set' or speedMode == 'limit' or speedMode == 'legal' or speedMode == 'off' then
    M.speedMode = speedMode
  else
    M.speedMode = nil
  end
end

local function resetSpeedModeAndValue()
  M.speedMode = nil -- maybe this should be 'off'
  M.routeSpeed = nil
end

local function setAggressionInternal(v)
  aggression = v and v or M.extAggression
end

local function setAggressionExternal(v)
  M.extAggression = v or M.extAggression
  setAggressionInternal()
  stateChanged()
end

local function setAggressionMode(aggrmode)
  if aggrmode == 'rubberBand' then
    aggressionMode = 'rubberBand'
  else
    aggressionMode = nil
  end
end

local function resetAggression()
  setAggressionInternal()
end

local function setForceLateral(v)
  forceLateral = v or 0
end

local function setDistPastEdge(v)
  distPastEdge = v or 0
end

local function setTargetObjectID(id)
  M.targetObjectID = M.targetObjectID ~= objectId and id or -1
  if M.targetObjectID ~= -1 then targetObjectSelectionMode = 'manual' end
end

local function updatePlayerData()
  if mapmgr.objects[M.targetObjectID] and targetObjectSelectionMode == 'manual' then
    player = mapmgr.objects[M.targetObjectID]
    player.id = M.targetObjectID
  elseif tableSize(mapmgr.objects) == 2 then
    if player ~= nil then
      player = mapmgr.objects[player.id]
    else
      for k, v in pairs(mapmgr.objects) do
        if k ~= objectId then
          M.targetObjectID = k
          player = v
          break
        end
      end
      targetObjectSelectionMode = 'auto'
    end
  else
    if player ~= nil and player.active == true then
      player = mapmgr.objects[player.id]
    else
      for k, v in pairs(mapmgr.objects) do
        if k ~= objectId and v.active == true then
          M.targetObjectID = k
          player = v
          break
        end
      end
      targetObjectSelectionMode = 'targetActive'
    end
  end
  mapmgr.objects[objectId] = mapmgr.objects[objectId] or {pos = aiPos, dirVec = aiDirVec}
end

local function driveCar(steering, throttle, brake, parkingbrake)
  input.event("steering", steering, 1)
  input.event("throttle", throttle, 2)
  input.event("brake", brake, 2)
  input.event("parkingbrake", parkingbrake, 2)

  lastCommand.steering = steering
  lastCommand.throttle = throttle
  lastCommand.brake = brake
  lastCommand.parkingbrake = parkingbrake
end

local function driveToTarget(targetPos, throttle, brake, targetSpeed)
  if not targetPos then return end

  local brakeCoef = 1
  local throttleCoef = 1

  local targetVec = (targetPos - aiPos):normalized()

  local dirAngle = math.asin(aiRightVec:dot(targetVec))
  local dirVel = aiVel:dot(aiDirVec)
  local absAiSpeed = abs(dirVel)
  local plan = currentRoute and currentRoute.plan
  targetSpeed = targetSpeed or plan and plan.targetSpeed
  if not targetSpeed then return end

  -- oversteer
  if aiSpeed > 1 then
    local rightVel = aiRightVec:dot(aiVel)
    if rightVel * aiRightVec:dot(targetPos - aiPos) > 0 then
      local rotVel = min(1, (prevDirVec:projectToOriginPlane(aiUpVec):normalized() - aiDirVec):length() * dt * 10000)
      throttleCoef = throttleCoef * max(0, 1 - abs(rightVel * aiSpeed * 0.05) * min(1, dirAngle * dirAngle * aiSpeed * 6) * rotVel)
    end
  end

  if plan and plan[3] and dirVel > 3 then
    local p1, p2 = plan[1].pos, plan[2].pos
    local p2p1 = p2 - p1
    local turnRight = p2p1:cross(aiUpVec):normalized()
    local tp2
    if plan.targetSeg and plan.targetSeg > 1 then
      tp2 = (targetPos - p2):normalized():dot(turnRight)
    else
      tp2 = (plan[3].pos - p2):normalized():dot(turnRight)
    end

    local outDeviation = aiDeviationSmoother:value() - aiDeviation * sign(tp2)
    outDeviation = sign(outDeviation) * min(1, abs(outDeviation))
    aiDeviationSmoother:set(outDeviation)
    aiDeviationSmoother:getUncapped(0, dt)

    if outDeviation > 0 and absAiSpeed > 3 then
      local steerCoef = outDeviation * absAiSpeed * absAiSpeed * min(1, dirAngle * dirAngle * 4)
      local understeerCoef = max(0, steerCoef) * min(1, abs(aiVel:dot(p2p1:normalized()) * 3))
      local noUndersteerCoef = max(0, 1 - understeerCoef)
      throttleCoef = throttleCoef * noUndersteerCoef
      brakeCoef = min(brakeCoef, max(0, 1 - understeerCoef * understeerCoef))
    end
  else
    aiDeviationSmoother:set(0)
  end

  -- wheel speed
  if absAiSpeed > 0.2 then
    local avgWheelSpeed = 0
    local wheelCount = 0
    local minAbsWheelSpeed = math.huge
    local maxAbsWheelSpeed = 0
    local lwheels = wheels.wheels
    for i = 0, tableSizeC(lwheels) - 1 do
      local wd = lwheels[i]
      if not wd.isBroken then
        wheelCount = wheelCount + wd.isSpeedo
        local absWheelVel = abs(wd.angularVelocity * wd.radius)
        avgWheelSpeed = avgWheelSpeed + absWheelVel * wd.isSpeedo
        if wd.brakeTorque > 0 then
          minAbsWheelSpeed = min(minAbsWheelSpeed, absWheelVel)
        end
        if wd.isPropulsed then
          maxAbsWheelSpeed = max(maxAbsWheelSpeed, absWheelVel)
        end
      end
    end

    -- manual abs and tcs
    if wheelCount > 0 then
      local latVgrad = 0.5 * aiWidth * abs(sensors.gx2) / (aiSpeed + 1e-30)
      avgWheelSpeed = avgWheelSpeed / wheelCount

      -- abs
      local minGradSpeed = absAiSpeed - latVgrad
      if avgWheelSpeed < minGradSpeed and brake > 0 then
        brakeCoef = brakeCoef * minAbsWheelSpeed / minGradSpeed
      end

      -- tcs
      local maxGradSpeed = max(3, absAiSpeed + latVgrad)
      if maxAbsWheelSpeed > maxGradSpeed and throttle > 0 then
        local redcoef = maxGradSpeed / maxAbsWheelSpeed
        throttleCoef = throttleCoef * redcoef * redcoef
      end
    end
  end

  local dirTarget = aiDirVec:dot(targetVec)
  local dirDiff = math.asin(aiDirVec:cross(aiUpVec):normalized():dot(targetVec))

  if crashManeuver == 1 and dirTarget < aiDirVec:dot(crashDir) then
    driveCar(-fsign(dirDiff), brake * brakeCoef, throttle * throttleCoef, 0)
    return
  else
    crashManeuver = 0
  end

  aiForceGoFrontTime = max(0, aiForceGoFrontTime - dt)
  if threewayturn.state == 1 and aiCannotMoveTime > 1 and aiForceGoFrontTime == 0 then
    threewayturn.state = 0
    aiCannotMoveTime = 0
    aiForceGoFrontTime = 2
  end

  if aiForceGoFrontTime > 0 and dirTarget < 0 then
    dirTarget = -dirTarget
    dirDiff = -dirDiff
  end

  if (dirTarget < 0 or (dirTarget < 0.15 and threewayturn.state == 1)) and currentRoute and not damageFlag then
    local n1, n2, n3 = plan[1], plan[2], plan[3]
    local edgeDist = min((n2 or n1).radiusOrig, n1.radiusOrig) - aiPos:z0():distanceToLine((n3 or n2).posOrig:z0(), n2.posOrig:z0())
    if edgeDist > aiWidth and threewayturn.state == 0 then
      driveCar(fsign(dirDiff), 0.5 * throttleCoef, 0, min(max(aiSpeed - 3, 0), 1))
    else
      if threewayturn.state == 0 then
        threewayturn.state = 1
        threewayturn.speedDifInt = 0
      end
      local angleModulation = min(max(0, -(dirTarget-0.15)), 1)
      local speedDif = (10 * aggression * angleModulation) - aiSpeed
      threewayturn.speedDifInt = threewayturn.speedDifInt + speedDif * dt
      local pbrake = clamp(sign2(aiDirVec:dot(gravityDir) - 0.17), 0, 1) -- apply parking brake if reversing on an incline >~ 10 deg
      driveCar(-sign2(dirDiff), 0, clamp(0.05 * speedDif + 0.01 * threewayturn.speedDifInt, 0, 1), pbrake)
    end
  else
    threewayturn.state = 0
    local pbrake
    if aiVel:dot(aiDirVec) * max(aiSpeed*aiSpeed - 1e-2, 0) < 0 then
      if aiSpeed < 0.15 and targetSpeed <= 1e-5 then
        pbrake = 1
      else
        pbrake = 0
      end
      driveCar(dirDiff, 0.5 * throttleCoef, 0, pbrake)
    else
      if (aiSpeed > 4 and aiSpeed < 30 and abs(dirDiff) > 0.97 and brake == 0) or (aiSpeed < 0.15 and targetSpeed <= 1e-5) then
        pbrake = 1
      else
        pbrake = 0
      end
      driveCar(dirDiff, throttle * throttleCoef, brake * brakeCoef, pbrake)
    end
  end
end

local function posOnPlan(pos, plan, dist)
  if not plan then return end
  dist = dist or 4
  dist = dist * dist
  local bestSeg, bestXnorm, bestDist
  for i = 1, #plan-2 do
    local p0, p1 = plan[i].pos, plan[i+1].pos
    local xnorm1 = pos:xnormOnLine(p0, p1)
    if xnorm1 > 0 then
      local p2 = plan[i+2].pos
      local xnorm2 = pos:xnormOnLine(p1, p2)
      if xnorm1 < 1 then -- contained in segment i
        if xnorm2 > 0 then -- also partly contained in segment i+1
          local sqDistFromP1 = pos:squaredDistance(p1)
          if sqDistFromP1 <= dist then
            bestSeg = i
            bestXnorm = 1
            break -- break inside conditional
          end
        else
          local sqDistFromLine = pos:squaredDistance(p0 + (p1 - p0) * xnorm1)
          if sqDistFromLine <= dist then
            bestSeg = i
            bestXnorm = xnorm1
          end
          break -- break should be outside above conditional
        end
      elseif xnorm2 < 0 then
        local sqDistFromP1 = pos:squaredDistance(p1)
        if sqDistFromP1 <= dist then
          bestSeg = i
          bestXnorm = 1
        end
        break -- break outside conditional
      end
    else
      break
    end
  end

  return bestSeg, bestXnorm
end

local function aiPosOnPlan(plan, planCount)
  local aiSeg = 1
  local aiXnormOnSeg = 0
  for i = 1, planCount-1 do
    local p0Pos, p1Pos = plan[i].pos, plan[i+1].pos
    local xnorm = aiPos:xnormOnLine(p0Pos, p1Pos)
    if xnorm < 1 then
      if i < planCount - 2 then
        local nextXnorm = aiPos:xnormOnLine(p1Pos, plan[i+2].pos)
        if nextXnorm >= 0 then
          local p1Radius = plan[i+1].radius
          if aiPos:squaredDistance(linePointFromXnorm(p1Pos, plan[i+2].pos, nextXnorm)) <
          square(aiWidth + lerp(p1Radius, plan[i+2].radius, min(1, nextXnorm))) then
            aiXnormOnSeg = nextXnorm
            aiSeg = i + 1
            break
          end
        end
      end
      aiXnormOnSeg = xnorm
      aiSeg = i
      break
    end
  end

  for _ = 1, aiSeg-1 do
    tableRemove(plan, 1)
    planCount = planCount-1
  end

  return aiXnormOnSeg, planCount
end

local function calculateTarget(plan, planCount)
  planCount = planCount or #plan
  local aiXnormOnSeg
  aiXnormOnSeg, planCount = aiPosOnPlan(plan, planCount)
  local targetLength = max(aiSpeed * 0.65, 4.5)

  if planCount >= 3 then
    local xnorm = clamp(aiXnormOnSeg, 0, 1)
    local p2Pos = plan[2].pos
    targetLength = max(targetLength, plan[1].pos:distance(p2Pos) * (1-xnorm), p2Pos:distance(plan[3].pos) * xnorm)
  end

  local remainder = targetLength

  local targetPos = plan[planCount].pos
  local targetSeg = max(1, planCount-1)
  local prevPos = linePointFromXnorm(plan[1].pos, plan[2].pos, aiXnormOnSeg) -- aiPos

  for i = 2, planCount do
    local pos = plan[i].pos
    local segVec = pos - prevPos
    local segLen = segVec:length()

    if remainder <= segLen then
      targetSeg = i - 1
      targetPos = prevPos + segVec * remainder / (segLen + 1e-25)

      -- smooth target
      local xnorm = clamp(targetPos:xnormOnLine(prevPos, pos), 0, 1)
      local lp_n1n2 = linePointFromXnorm(prevPos, pos, xnorm * 0.5 + 0.25)
      if xnorm <= 0.5 then
        if i >= 3 then
          targetPos = linePointFromXnorm(linePointFromXnorm(plan[i-2].pos, prevPos, xnorm * 0.5 + 0.75), lp_n1n2, xnorm + 0.5)
        end
      else
        if i <= planCount - 2 then
          targetPos = linePointFromXnorm(lp_n1n2, linePointFromXnorm(pos, plan[i+1].pos, xnorm * 0.5 - 0.25), xnorm - 0.5)
        end
      end
      break
    end
    prevPos = pos
    remainder = remainder - segLen
  end

  return targetPos, targetSeg, aiXnormOnSeg, planCount
end

local function targetsCompatible(baseRoute, newRoute)
  local baseTvec = baseRoute.plan.targetPos - aiPos
  local newTvec = newRoute.plan.targetPos - aiPos
  if aiSpeed < 2 then return true end
  if newTvec:dot(aiDirVec) * baseTvec:dot(aiDirVec) <= 0 then return false end
  local baseTargetRight = baseTvec:cross(aiUpVec):normalized()
  return abs(newTvec:normalized():dot(baseTargetRight)) * aiSpeed < 2
end

local function getMinPlanLen(limLow, speed, accelg)
  -- given current speed, distance required to come to a stop if I can decelerate at 0.2g
  limLow = limLow or 150
  speed = speed or aiSpeed
  accelg = max(0.2, accelg or 0.2)
  return min(550, max(limLow, 0.5 * speed * speed / (accelg * g)))
end

local function pickAiWp(wp1, wp2, dirVec)
  dirVec = dirVec or aiDirVec
  local vec1 = map.nodes[wp1].pos - aiPos
  local vec2 = map.nodes[wp2].pos - aiPos
  local dot1 = vec1:dot(dirVec)
  local dot2 = vec2:dot(dirVec)
  if (dot1 * dot2) <= 0 then
    if dot1 < 0 then
      return wp2
    end
  else
    if vec2:squaredLength() < vec1:squaredLength() then
      return wp2
    end
  end
  return wp1
end

local function pathExtend(path, newPath)
  if newPath == nil then return end
  local pathCount = #path
  if path[pathCount] ~= newPath[1] then return end
  pathCount = pathCount - 1
  for i = 2, #newPath do
    path[pathCount+i] = newPath[i]
  end
end

-- http://cnx.org/contents/--TzKjCB@8/Projectile-motion-on-an-inclin
local function projectileSqSpeedToRangeRatio(pos1, pos2, pos3)
  local sinTheta = (pos2.z - pos1.z) / pos1:distance(pos2)
  local sinAlpha = (pos3.z - pos2.z) / pos2:distance(pos3)
  local cosAlphaSquared = max(1 - sinAlpha * sinAlpha, 0)
  local cosTheta = sqrt(max(1 - sinTheta * sinTheta, 0)) -- in the interval theta = {-pi/2, pi/2} cosTheta is always positive
  return 0.5 * g * cosAlphaSquared / max(cosTheta * (sinTheta*sqrt(cosAlphaSquared) - cosTheta*sinAlpha), 0)
end

local function inCurvature(vec1, vec2)
  local vec1Sqlen, vec2Sqlen = vec1:squaredLength(), vec2:squaredLength()
  local dot12 = vec1:dot(vec2)
  local cos8sq = min(1, dot12 * dot12 / max(1e-30, vec1Sqlen * vec2Sqlen))

  if dot12 < 0 then
    local minDsq = min(vec1Sqlen, vec2Sqlen)
    local maxDsq = minDsq / max(1e-30, cos8sq)
    if max(vec1Sqlen, vec2Sqlen) > (minDsq + maxDsq) * 0.5 then
      if vec1Sqlen > vec2Sqlen then
        vec1, vec2 = vec2, vec1
        vec1Sqlen, vec2Sqlen = vec2Sqlen, vec1Sqlen
      end
      vec2 = sqrt(0.5 * (minDsq + maxDsq) / max(1e-30, vec2Sqlen)) * vec2
    end
  end

  return 2 * sqrt((1 - cos8sq) / max(1e-30, (vec1 + vec2):squaredLength()))
end

local function getPathLen(path, startIdx, stopIdx)
  if not path then return end
  startIdx = startIdx or 1
  stopIdx = stopIdx or #path
  local mapNodes = map.nodes
  local pathLen = 0
  for i = startIdx+1, stopIdx do
    pathLen = pathLen + (mapNodes[path[i]].pos - mapNodes[path[i-1]].pos):length()
  end

  return pathLen
end

local function waypointInPath(path, waypoint, startIdx, stopIdx)
  if not path or not waypoint then return end
  startIdx = startIdx or 1
  stopIdx = stopIdx or #path
  for i = startIdx, stopIdx do
    if path[i] == waypoint then
      return i
    end
  end
end

local function getPlanLen(plan, planCount)
  planCount = planCount or #plan
  local planLen = 0
  for i = 2, planCount do
    planLen = planLen + (plan[i].pos - plan[i-1].pos):length()
  end
  return planLen, planCount
end

local function buildNextRoute(plan, planCount, path)
  local nextPathIdx = (plan[planCount].pathidx or 0) + 1 -- if the first plan node is the aiPos it does not have a pathIdx value yet

  if race == true and noOfLaps and noOfLaps > 1 and not path[nextPathIdx] then -- in case the path loops
    local loopPathId
    local pathCount = #path
    local lastWayPoint = path[pathCount]
    for i = 1, pathCount do
      if lastWayPoint == path[i] then
        loopPathId = i
        break
      end
    end
    nextPathIdx = 1 + loopPathId -- nextPathIdx % #path
    noOfLaps = noOfLaps - 1
  end

  local nextNodeName = path[nextPathIdx]
  if not nextNodeName then return end

  local n1, oneWay
  if path[nextPathIdx-1] then
    n1 = map.nodes[path[nextPathIdx-1]] -- this might be a bug, try this -> route.plan[planCount].posOrig
    local link = n1.links[path[nextPathIdx]]
    local legalSpeed
    if link then
      oneWay = link.oneWay
      if nextPathIdx > 2 then
        legalSpeed = min(link.speedLimit, n1.links[path[nextPathIdx-2]].speedLimit)
      else
        legalSpeed = link.speedLimit
      end
    end
    plan[planCount].legalSpeed = legalSpeed
  elseif path[nextPathIdx+1] then -- if previous plan node is the aiPos then get oneWay data from path segment ahead of nextPathIdx
    n1 = {pos = vec3(aiPos), radius = aiWidth * 0.5}
    local link = map.nodes[path[nextPathIdx]].links[path[nextPathIdx+1]]
    oneWay = link and link.oneWay
  end

  local n2 = map.nodes[nextNodeName]
  if not n2 then return end

  if driveInLaneFlag then
    local lane = 1
    plan.lane = plan.lane or 0
    if oneWay then
      if plan.lane ~= 0 then
        lane = plan.lane
      else
        if path[2] then
          local curPathIdx = (plan[1] and plan[1].pathidx) and max(2, plan[1].pathidx) or 2
          local p1Pos = map.nodes[path[curPathIdx-1]].pos
          lane = sign((map.nodes[path[curPathIdx]].pos - p1Pos):z0():cross(gravityDir):dot(p1Pos - aiPos))
          plan.lane = lane
        end
      end
    else
      plan.lane = 0
    end

    local n2Pos = n2.pos
    local nVec1
    if path[nextPathIdx-1] then
      nVec1 = (n1.pos - n2Pos):z0():cross(gravityDir):normalized()
    else
      nVec1 = vec3()
    end

    local nVec2
    if path[nextPathIdx+1] then
      nVec2 = (n2Pos - map.nodes[path[nextPathIdx+1]].pos):z0():cross(gravityDir):normalized()
    else
      nVec2 = vec3()
    end

    local width = max(n2.radius*0.5, aiWidth*0.7)
    local displacement = max(0, n2.radius - width) -- provide a bit more space in narrow roads so other vehicles can overtake
    n2 = {pos = n2Pos + displacement * lane * (1 - nVec1:dot(nVec2) * 0.5) * (nVec1 + nVec2), radius = width}
  end

  local lastPlanPos = plan[planCount] and plan[planCount].pos or aiPos
  local vec = (lastPlanPos - n2.pos):z0()
  local manSpeed = speedList and speedList[nextNodeName]

  return {pos = vec3(n2.pos), posOrig = vec3(n2.pos), radius = n2.radius, radiusOrig = n2.radius,
          posz0 = n2.pos:z0(), vec = vec, dirVec = vec:normalized(), turnDir = vec3(0,0,0),
          manSpeed = manSpeed, pathidx = nextPathIdx}
end

local function mergePathPrefix(source, dest, srcStart)
  srcStart = srcStart or 1
  local sourceCount = #source
  local dict = table.new(0, sourceCount-(srcStart-1))
  for i = srcStart, sourceCount do
    dict[source[i]] = i
  end

  local destCount = #dest
  for i = destCount, 1, -1 do
    local srci = dict[dest[i]]
    if srci ~= nil then
      local res = table.new(destCount, 0)
      local resi = 1
      for i1 = srcStart, srci - 1 do
        res[resi] = source[i1]
        resi = resi + 1
      end
      for i1 = i, destCount do
        res[resi] = dest[i1]
        resi = resi + 1
      end

      return res, srci
    end
  end

  return dest, 0
end

local function planAhead(route, baseRoute)
  if route == nil then return end
  if route.path == nil then
    route.path = {}
    for i = 1, #route do
      route.path[i] = route[i]
      route[i] = nil
    end
    route.plan = {}
  end

  local plan = route.plan
  local planCount

  if baseRoute and not plan[1] then
    planCount = 0
    -- merge from base plan
    local bsrPlan = baseRoute.plan
    if bsrPlan[2] then
      local commonPathEnd
      route.path, commonPathEnd = mergePathPrefix(baseRoute.path, route.path, bsrPlan[2].pathidx)
      if commonPathEnd >= 1 then
        local refpathidx = bsrPlan[2].pathidx - 1
        for i = 1, #bsrPlan do
          local n = bsrPlan[i]
          if n.pathidx > commonPathEnd then break end
          plan[i] = {pos = vec3(n.pos), posOrig = vec3(n.posOrig), posz0 = vec3(n.posz0),
                    vec = vec3(n.vec), dirVec = vec3(n.dirVec), turnDir = vec3(n.turnDir),
                    radius = n.radius, radiusOrig = n.radiusOrig, pathidx = max(1, n.pathidx-refpathidx), legalSpeed = n.legalSpeed}
          planCount = i
        end
        if planCount > bsrPlan.targetSeg then
          plan.targetSeg = bsrPlan.targetSeg
          plan.targetPos = vec3(bsrPlan.targetPos)
          plan.aiSeg = bsrPlan.aiSeg
        end
      end
    end
  end

  if not plan[1] then
    plan[1] = {posOrig = vec3(aiPos), pos = vec3(aiPos), posz0 = aiPos:z0(), vec = 8 * vec3(-aiDirVec),
               dirVec = vec3(-aiDirVec), turnDir = vec3(0,0,0), radiusOrig = 2, radius = 2}
    planCount = 1
  end

  local planLen
  planLen, planCount = getPlanLen(plan, planCount)
  local minPlanLen = getMinPlanLen()
  while not plan[minPlanCount] or planLen < minPlanLen do
    local n = buildNextRoute(plan, planCount, route.path)
    if not n then break end
    planCount = planCount + 1
    plan[planCount] = n
    planLen = planLen + (n.pos-plan[planCount-1].pos):length()
  end

  if not plan[2] then return end
  if not plan[1].pathidx then plan[1].pathidx = plan[2].pathidx end

  do
    local segmentSplitDelay = plan.segmentSplitDelay or 0
    local distOnPlan = 0
    for i = 1, planCount-1 do
      local curDist = plan[i].posOrig:squaredDistance(plan[i+1].posOrig)
      local xSq = square(distOnPlan)
      if curDist > square(min(220, (25e-8*xSq + 1e-5)*xSq + 6)) and distOnPlan < 550 then
        if segmentSplitDelay == 0 then
          local n1, n2 = plan[i], plan[i+1]
          local pos = (n1.pos + n2.pos) * 0.5
          local vec = (n1.pos - pos):z0()
          n2.vec = (pos - n2.pos):z0()
          n2.dirVec = n2.vec:normalized()
          local legalSpeed
          if n2.pathidx > 1 then
            legalSpeed = map.nodes[route.path[n2.pathidx]].links[route.path[n2.pathidx-1]].speedLimit
          else
            legalSpeed = n2.legalSpeed
          end
          tableInsert(plan, i+1, {posOrig = (n1.posOrig + n2.posOrig)*0.5, pos = pos, posz0 = pos:z0(),
                                  vec = vec, dirVec = vec:normalized(), turnDir = vec3(0, 0, 0),
                                  radiusOrig = (n1.radiusOrig + n2.radiusOrig)*0.5, radius = (n1.radius + n2.radius)*0.5,
                                  pathidx = n2.pathidx, legalSpeed = legalSpeed})
          planCount = planCount + 1
          segmentSplitDelay = min(5, floor(90/aiSpeed))
        else
          segmentSplitDelay = segmentSplitDelay - 1
        end
        break
      end
      distOnPlan = distOnPlan + sqrt(curDist)
    end
    plan.segmentSplitDelay = segmentSplitDelay
  end

  if plan.targetSeg == nil then
    local aiXnormOnSeg
    plan.targetPos, plan.targetSeg, aiXnormOnSeg, planCount = calculateTarget(plan, planCount)
    plan.aiSeg = 1
  end

  for i = 0, planCount do
    if forces[i] then
      forces[i]:set(0,0,0)
    else
      forces[i] = vec3(0,0,0)
    end
  end

  -- apply springs forces
  for i = 1, planCount-1 do
    local n1 = plan[i]
    local n2 = plan[i+1]
    local v1 = n1.dirVec
    local v2 = -n2.dirVec
    local turnDir = (v1 + v2):normalized()

    local lforce = v1:cross(gravityVec):normalized() * forceLateral * math.pow(abs(v1:dot(v2)), 4)
    local nforce = max(1 + v1:dot(v2), 0) * 5 * turnDir
    forces[i+1] = forces[i+1] - nforce - lforce
    forces[i-1] = forces[i-1] - nforce - lforce
    forces[i] = forces[i] + nforce * 2 + lforce * 2

    n1.turnDir:set(turnDir)
    n1.speed = 0
  end

  -- other vehicle awareness
  plan.trafficMinProjSpeed = math.huge
  plan.trafficBlockCoef = nil
  if avoidCars ~= 'off' then
    table.clear(trafficTable)
    local trafficTableLen = 0
    for plID, v in pairs(mapmgr.objects) do
      if plID ~= objectId and (M.mode ~= 'chase' or plID ~= player.id or chaseData.suspectState == 'stopped') then
        v.length = obj:getObjectInitialLength(plID) + 0.3
        v.width = obj:getObjectInitialWidth(plID)
        v.targetType = (player and plID == player.id) and M.mode
        if v.targetType == 'follow' then
          v.width = v.width * 4
        end
        local posFront = obj:getObjectFrontPosition(plID)
        local dirVec = v.dirVec
        v.posFront = dirVec * 0.3 + posFront
        v.posRear = dirVec * (-v.length) + posFront
        v.lightbar = ((v.states and v.states.lightbar) and v.states.lightbar > 0) and true or false
        table.insert(trafficTable, v)
        trafficTableLen = trafficTableLen + 1
      end
    end

    local trafficMinSpeedSq = math.huge
    local distanceT = 0
    local aiPathVel = aiVel:dot((plan[2].pos-plan[1].pos):normalized())
    local aiPathVelInv = 1 / abs(aiPathVel + 1e-30)
    local minTrafficDir = 1

    for i = 2, planCount-1 do
      local n1, n2 = plan[i], plan[i+1]
      local n1pos, n2pos = n1.pos, n2.pos
      local n1n2 = n2pos - n1pos
      local n1n2len = n1n2:length()
      local nDir = n1n2 / (n1n2len + 1e-30)
      n1.trafficSqVel = math.huge
      local arrivalT = distanceT * aiPathVelInv

      if damageFlag then
        n1.trafficSqVel = 0
      else
        for j = trafficTableLen, 1, -1 do
          local v = trafficTable[j]
          local plPosFront, plPosRear, plWidth = v.posFront, v.posRear, v.width
          local ai2PlVec = plPosFront - aiPos
          local ai2PlDir = ai2PlVec:dot(aiDirVec)
          local ai2PlSqDist = ai2PlVec:squaredLength()
          if ai2PlDir > 0 then
            local velDisp = arrivalT * v.vel
            plPosFront = plPosFront + velDisp
            plPosRear = plPosRear + velDisp
          end
          local extVec = nDir * (max(aiWidth, plWidth) * 0.5)
          local n1ext, n2ext = n1pos - extVec, n2pos + extVec
          local rnorm, vnorm = closestLinePoints(n1ext, n2ext, plPosFront, plPosRear)

          if M.mode == 'traffic' and v.lightbar and ai2PlSqDist <= 6400 then -- lightbar awareness
            forces[i] = forces[i] + aiRightVec * 3
            forces[i + 1] = forces[i + 1] + aiRightVec * 3
            trafficSide.cTimer = max(2.5, trafficSide.cTimer)
            n1.trafficSqVel = clamp((ai2PlSqDist / 15) - 16, 0, 64)
          end

          local minSqDist = math.huge
          if rnorm > 0 and rnorm < 1 and vnorm > 0 and vnorm < 1 then
            minSqDist = 0
          else
            local rlen = n1n2len + plWidth
            local xnorm = plPosFront:xnormOnLine(n1ext, n2ext) * rlen
            if xnorm > 0 and xnorm < rlen then minSqDist = min(minSqDist, (n1ext+nDir*xnorm-plPosFront):squaredLength()) end

            xnorm = plPosRear:xnormOnLine(n1ext, n2ext) * rlen
            if xnorm > 0 and xnorm < rlen then minSqDist = min(minSqDist, (n1ext+nDir*xnorm-plPosRear):squaredLength()) end

            rlen = v.length + aiWidth
            local v1 = n1ext - plPosRear
            local v1dot = v1:dot(v.dirVec)
            if (v1dot > 0 and v1dot < rlen) then minSqDist = min(minSqDist, (v1 - v1dot * v.dirVec):squaredLength()) end

            v1 = n2ext - plPosRear
            v1dot = v1:dot(v.dirVec)
            if (v1dot > 0 and v1dot < rlen) then minSqDist = min(minSqDist, (v1 - v1dot * v.dirVec):squaredLength()) end
          end

          if minSqDist < square((aiWidth + plWidth) * 0.8) then
            local velProjOnSeg = max(0, v.vel:dot(nDir))
            local middlePos = (plPosFront + plPosRear) * 0.5
            local forceCoef = trafficSide.side * 0.5 * max(max(0, aiSpeed - velProjOnSeg), max(0, trafficSide.cTimer - 8) * 0.25) / ((1 + minSqDist) * (1 + distanceT * min(0.1, 1 / (2 * max(0, aiPathVel - v.vel:dot(nDir)) + 1e-30))))

            if v.targetType ~= 'follow' then
              forces[i] = forces[i] - (sign(n1.turnDir:dot(middlePos - n1.posOrig)) * forceCoef) * n1.turnDir
              forces[i+1] = forces[i+1] - (sign(n2.turnDir:dot(middlePos - n2.posOrig)) * forceCoef) * n2.turnDir
            end

            if avoidCars == 'on' and M.mode ~= 'flee' and M.mode ~= 'random' then
              if minSqDist < square((aiWidth + plWidth) * 0.51)  then
                -- obj.debugDrawProxy:drawSphere(0.25, v.posFront:toFloat3(), color(0,0,255,255))
                -- obj.debugDrawProxy:drawSphere(0.25, plPosFront:toFloat3(), color(0,0,255,255))
                table.remove(trafficTable, j)
                trafficTableLen = trafficTableLen - 1
                plan.trafficMinProjSpeed = min(plan.trafficMinProjSpeed, velProjOnSeg)

                n1.trafficSqVel = min(n1.trafficSqVel, velProjOnSeg * velProjOnSeg)
                trafficMinSpeedSq = min(trafficMinSpeedSq, v.vel:squaredLength())
                minTrafficDir = min(minTrafficDir, v.dirVec:dot(nDir))
              end

              if i == 2 and minSqDist < square((aiWidth + plWidth) * 0.6) and ai2PlDir > 0 and v.vel:dot(aiRightVec) * ai2PlVec:dot(aiRightVec) < 0 then
                n1.trafficSqVel = max(0, n1.trafficSqVel - (v.vel - v.vel:dot(aiDirVec) * v.vel):length())
              end
            end
          end
        end
      end
      distanceT = distanceT + n1n2len
    end

    if max(trafficMinSpeedSq, aiSpeed*aiSpeed) < 0.25 then
      plan.trafficBlockCoef = clamp((1 - minTrafficDir) * 0.5, 0, 1)
    else
      plan.trafficBlockCoef = 0
    end
    plan[1].trafficSqVel = plan[2].trafficSqVel
  end

  -- spring force integrator
  local distFromEdge = aiWidth * min(1, 0.7 + 0.3 * square(max(0, 1 - aggression)))
  for i = 2, planCount do
    local n = plan[i]
    local k = n.turnDir:dot(forces[i])
    local nodeDisplVec = n.pos + fsign(k) * min(abs(k), 0.05) * n.turnDir - n.posOrig
    local nodeDisplLen = nodeDisplVec:length()
    local nodeDisplLenLim = min(nodeDisplLen, max(n.radiusOrig - distFromEdge + distPastEdge, 0))
    --n.radius = max(n.radiusOrig - fsign(nodeDisplVec:dot(n.turnDir)) * nodeDisplLenLim - distFromEdge, distFromEdge)
    n.radius = max(n.radiusOrig - nodeDisplLenLim, aiWidth * 0.7)
    n.pos = n.posOrig + (nodeDisplLenLim / (nodeDisplLen + 1e-30)) * nodeDisplVec
    n.posz0:set(n.pos:z0())
    n.vec:set(plan[i-1].posz0 - n.posz0)
    n.dirVec:set(n.vec:normalized())
  end

  local targetSeg = plan.targetSeg
  -- smoothly distribute error from planline onto the front segments
  if targetSeg ~= nil and planCount > targetSeg and plan.targetPos and threewayturn.state == 0 then
    local dTotal = 0
    local sumLen = {}
    for i = 2, targetSeg - 1  do
      sumLen[i] = dTotal
      dTotal = dTotal + (plan[i+1].pos - plan[i].pos):length()
    end
    dTotal = dTotal + (plan.targetPos - plan[targetSeg].pos):length()

    dTotal = max(1, dTotal)
    local p1, p2 = plan[1].pos, plan[2].pos
    local dispVec = (aiPos - linePointFromXnorm(p1, p2, aiPos:xnormOnLine(p1, p2))) * (0.5*dt)
    aiDeviation = dispVec:dot((p2-p1):cross(aiUpVec):normalized())

    local dispVecRatio = dispVec / dTotal
    for i = targetSeg - 1, 3, -1 do
      plan[i].pos = plan[i].pos + (dTotal - sumLen[i]) * dispVecRatio
      plan[i].posz0 = plan[i].pos:z0()
      plan[i+1].vec = plan[i].posz0 - plan[i+1].posz0
      plan[i+1].dirVec = plan[i+1].vec:normalized()
    end

    plan[1].pos = p1 + dispVec
    plan[1].posz0 = plan[1].pos:z0()
    plan[2].pos = p2 + dispVec
    plan[2].posz0 = plan[2].pos:z0()
    if plan[3] then
      plan[3].vec = plan[2].posz0 - plan[3].posz0
      plan[3].dirVec = plan[3].vec:normalized()
    end
  end

  plan.targetPos, plan.targetSeg, plan.aiXnormOnSeg, planCount = calculateTarget(plan)
  plan.aiSeg = 1
  plan.planCount = planCount

  -- plan speeds
  local totalAccel = min(aggression, staticFrictionCoef) * g

  local rLast = plan[planCount]
  if route.path[rLast.pathidx+1] or (race and noOfLaps and noOfLaps > 1) then
    rLast.speed = rLast.manSpeed or sqrt(2 * 550 * totalAccel) -- shouldn't this be calculated based on the path length remaining?
  else
    rLast.speed = rLast.manSpeed or 0
  end

  -- speed planning
  for i = planCount-1, 1, -1 do
    local n1 = plan[i]
    local n2 = plan[i+1]

    -- inclination calculation
    local v12d = n2.pos - n1.pos
    local dist = v12d:length() + 1e-30
    v12d = v12d / dist
    local Gf = gravityVec:dot(v12d) -- acceleration due to gravity parallel to road segment, positive when downhill
    local Gt = (gravityVec - v12d * Gf):length() / g -- gravity vec normal to road segment
    local Gn = gravityVec:normalized():dot(v12d) -- normalized gravity to road value

    local n2SpeedSq = square(n2.speed)

    local n0vec = plan[max(1, i-2)].posz0 - n1.posz0
    local n3vec = n1.posz0 - plan[min(planCount, i + 2)].posz0
    local curvature = min(
            min(inCurvature(n1.vec, n2.vec), inCurvature(n0vec, n3vec)),
            min(inCurvature(n0vec, n2.vec), inCurvature(n1.vec, n3vec))) + 1.6e-7

    local turnSpeedSq = (totalAccel * (1 - max(0, Gn))) * Gt / curvature -- available centripetal acceleration * radius

    -- https://physics.stackexchange.com/questions/312569/non-uniform-circular-motion-velocity-optimization
    --local deltaPhi = 2 * asin(0.5 * n2.vec:length() * curvature) -- phi = phi2 - phi1 = 2 * asin(halfcord / radius)
    local n1SpeedSq = turnSpeedSq * sin(min(asin(min(1, n2SpeedSq / turnSpeedSq)) + 2 * curvature * dist, pi * 0.5))

    n1SpeedSq = min(n1SpeedSq, n1.trafficSqVel or math.huge)
    n1.trafficSqVel = math.huge

    n1.speed = n1.manSpeed or
                (M.speedMode == 'legal' and n1.legalSpeed and min(n1.legalSpeed, sqrt(n1SpeedSq))) or
                (M.speedMode == 'limit' and M.routeSpeed and min(M.routeSpeed, sqrt(n1SpeedSq))) or
                (M.speedMode == 'set' and M.routeSpeed) or
                sqrt(n1SpeedSq)
  end

  plan.targetSpeed = plan[1].speed + max(0, plan.aiXnormOnSeg) * (plan[2].speed - plan[1].speed)

  return plan
end

local function doTrafficActions()
  -- horn
  if not hornFlag then
    if not damageFlag and trafficBlock.timer >= trafficBlock.horn then
      electrics.horn(true)
      hornFlag = true
    end
  else
    if trafficBlock.timer < trafficBlock.horn then
      electrics.horn(false)
      hornFlag = false
    end
  end

  -- hazard lights
  if mapmgr.objects[objectId] and mapmgr.objects[objectId].damage >= 1000 then
    if not damageFlag then
      trafficSignal.hTimer = trafficSignal.hTimer + dt
      if trafficSignal.hTimer > trafficSignal.hLimit then
        electrics.set_warn_signal(1)
        trafficSignal.hTimer = 0
        damageFlag = true
      end
    end
    return
  else
    damageFlag = false
  end

  -- intersections & turn signals
  local path = currentRoute.path
  local plan = currentRoute.plan

  if not intersection.pos or intersection.turn == 0 then
    for i = plan[1].pathidx, min(#path - 2, plan[1].pathidx + 3) do
      local nid2 = path[i + 1]
      local n2 = map.nodes[nid2]

      if tableSize(n2.links) > 2 then -- intersection
        local nid1 = path[i]
        local n1 = map.nodes[nid1]
        local nDir = (n2.pos - n1.pos):z0():normalized()
        local prevLinkTest = not (tableSize(n1.links) > 2 and n1.pos:squaredDistance(n2.pos) < square(n1.radius * 4)) -- previous branch within proximity

        for k, v in pairs(n2.links) do
          if k ~= nid1 then -- not previous node
            local linkDir = (map.nodes[k].pos - n2.pos):z0():normalized()
            local linkTurn = nDir:dot(linkDir)
            local drivabilityDiff = v.drivability - n1.links[nid2].drivability
            local radiusDiff = map.nodes[k].radius - n1.radius
            local oneWayTest = not (v.oneWay and v.inNode ~= k) -- one way outbound link

            if k == path[i + 2] and linkTurn < 0.6 then -- if current path goes to this link
              intersection.turn = -sign2(nDir:cross(gravityDir):dot(linkDir))
            end
            -- if turn exists, or if links conditions are met
            if intersection.turn ~= 0 or (not intersection.pos and oneWayTest and prevLinkTest and linkTurn < 0.8 and linkTurn > -0.6 and
            (drivabilityDiff > 0 or radiusDiff >= 0.5)) then
              intersection.pos = n2.pos - (n2.pos - n1.pos):normalized() * math.max(4, n2.radius) * 1.5 -- stop point
            end
          end
        end
      end
      if intersection.turn ~= 0 then break end
    end
  end

  if intersection.pos then
    local bestDist = math.huge
    local bestIdx
    local distSq = aiPos:squaredDistance(intersection.pos)

    for i = 1, #plan - 1 do
      if plan[i].manSpeed == 0 then plan[i].manSpeed = nil end
      local d = currentRoute.plan[i].pos:squaredDistance(intersection.pos)
      if d < bestDist then
        bestDist = d
        bestIdx = i
      end
    end

    if intersection.timer == 0 and distSq <= square(aiSpeed * 2) then -- approaching intersection
      if intersection.turn < 0 then
        if electrics.values.turnsignal >= 0 then
          electrics.toggle_left_signal()
        end
      elseif intersection.turn > 0 then
        if electrics.values.turnsignal <= 0 then
          electrics.toggle_right_signal()
        end
      end
    end

    if bestIdx then
      plan[bestIdx].manSpeed = 0 -- plan node stop
      if bestIdx <= 2 then intersection.timer = intersection.timer + dt end -- stopped at intersection
    end

    if intersection.timer > 3 then -- stop time
      plan[1].manSpeed, plan[2].manSpeed = nil, nil
      if distSq >= 400 then -- cleared intersection area (including other branch nodes)
        intersection.pos = nil
        intersection.timer = 0
        intersection.turn = 0
      end
    end
  end
end

local function resetMapAndRoute()
  map = nil
  currentRoute = nil
  race = nil
  noOfLaps = nil
  damageFlag = false
  internalState = 'onroad'
  changePlanTimer = 0
  resetAggression()
end

local function getMapEdges(cutOffDrivability, node)
  -- creates a table (edgeDict) with map edges with drivability > cutOffDrivability
  if map.nodes ~= nil then
    local allSCC = mapmgr.getSCC(node) -- An array of dicts containing all strongly connected components reachable from 'node'.
    local maxSccLen = 0
    local sccIdx
    for i, scc in ipairs(allSCC) do
      -- finds the scc with the most nodes
      local sccLen = scc[0] -- position at which the number of nodes in currentSCC is stored
      if sccLen > maxSccLen then
        sccIdx = i
        maxSccLen = sccLen
      end
      scc[0] = nil
    end
    local currentSCC = allSCC[sccIdx]
    local keySet = {}
    local keySetLen = 0
    edgeDict = {}
    for nid, n in pairs(map.nodes) do
      if currentSCC[nid] or not driveInLaneFlag then
        for lid, data in pairs(n.links) do
          if (currentSCC[lid] or not driveInLaneFlag) and (data.drivability > cutOffDrivability) then
            local inNode = data.inNode
            local outNode = inNode == nid and lid or nid
            keySetLen = keySetLen + 1
            keySet[keySetLen] = {inNode, outNode}
            edgeDict[inNode..'\0'..outNode] = 1
            if not data.oneWay or not driveInLaneFlag then
              edgeDict[outNode..'\0'..inNode] = 1
            end
          end
        end
      end
    end
    if keySetLen == 0 then return end
    local edge = keySet[math.random(keySetLen)]
    return edge[1], edge[2]
  end
end

local function newManualPath()
  local newRoute, n1, n2, dist
  local offRoad = false

  if manualPath then
    if currentRoute and currentRoute.path then
      pathExtend(currentRoute.path, manualPath)
    else
      newRoute = {plan = {}, path = manualPath}
      currentRoute = newRoute
    end
    manualPath = nil
  elseif wpList then
    if currentRoute and currentRoute.path then
      newRoute = {plan = currentRoute.plan, path = currentRoute.path}
    else
      n1, n2, dist = mapmgr.findClosestRoad(aiPos)

      if n1 == nil or n2 == nil then
        gui.message("Could not find a road network, or closest road is too far", 5, "AI debug")
        dump("Could not find a road network, or closest road is too far")
        return
      end

      if dist > 2 * max(map.nodes[n1].radius, map.nodes[n2].radius) then
        offRoad = true
        local vec1 = map.nodes[n1].pos - aiPos
        local vec2 = map.nodes[n2].pos - aiPos

        if aiDirVec:dot(vec1) > 0 and aiDirVec:dot(vec2) > 0 then
          if vec1:length() > vec2:length() then
            n1, n2 = n2, n1
          end
        elseif aiDirVec:dot(map.nodes[n2].pos - map.nodes[n1].pos) > 0 then
          n1, n2 = n2, n1
        end
      elseif aiDirVec:dot(map.nodes[n2].pos - map.nodes[n1].pos) > 0 then
        n1, n2 = n2, n1
      end

      newRoute = {plan = {}, path = {n1}}
    end

    for i = 0, #wpList-1 do
      local wp1 = wpList[i] or newRoute.path[#newRoute.path]
      local wp2 = wpList[i+1]
      local route = mapmgr.getPath(wp1, wp2, driveInLaneFlag and 10e7 or 1)
      local routeLen = #route
      if routeLen == 0 or (routeLen == 1 and wp2 ~= wp1) then
        gui.message("Path between waypoints '".. wp1 .."' - '".. wp2 .."' Not Found", 7, "AI debug")
        dump("Path between waypoints '".. wp1 .."' - '".. wp2 .."' Not Found")
        return
      end

      for j = 2, routeLen do
        tableInsert(newRoute.path, route[j])
      end
    end

    wpList = nil

    if not offRoad and newRoute.path[3] and newRoute.path[2] == n2 then
      tableRemove(newRoute.path, 1)
    end

    currentRoute = newRoute
  end
end

local function validateUserInput(list)
  validateInput = nop
  list = list or wpList
  if not list then return end
  --local isValid = wpList[1] and true or false
  local isValid = list[1] and true or false
  for i = 1, #list do -- #wpList
    if map.nodeAliases[list[i]] then
      local nodeAlias = map.nodeAliases[list[i]]
      if map.nodes[nodeAlias] then
        list[i] = nodeAlias
      else
        if isValid then
          gui.message("One or more of the waypoints were not found on the map. Check the game console for more info.", 6, "AI debug")
          dump('The waypoints with the following names could not be found on the Map')
          isValid = false
        end
        print(list[i])
      end
    end
  end

  return isValid
end

local function fleePlan()
  local newRoute
  if currentRoute and currentRoute.plan[2].pathidx > #currentRoute.path * 0.7 and currentRoute.path[3] and targetWPName == nil and internalState ~= 'offroad' and (aiPos - player.pos):dot(aiDirVec) >= 0 and currentRoute.plan.trafficMinProjSpeed > 3 then
    local path = currentRoute.path
    local pathCount = #path
    local cr1 = path[pathCount-1]
    local cr2 = path[pathCount]
    pathExtend(path, mapmgr.getFleePath(cr2, (map.nodes[cr2].pos - map.nodes[cr1].pos):normalized(), player.pos, getMinPlanLen(), 0.01, 0.01))
  elseif changePlanTimer == 0 then
    local wp1, wp2 = mapmgr.findClosestRoad(aiPos)
    if wp1 == nil or wp2 == nil then
      internalState = 'offroad'
      return
    else
      internalState = 'onroad'
    end

    local dirVec
    if currentRoute and currentRoute.plan.trafficMinProjSpeed < 3 then
      changePlanTimer = 5
      dirVec = -aiDirVec
    else
      dirVec = aiDirVec
    end

    local startnode = pickAiWp(wp1, wp2, dirVec)
    if not targetWPName then -- flee without target
      newRoute = mapmgr.getFleePath(startnode, dirVec, player.pos, getMinPlanLen(), 0.01, 0.01)
    else -- flee to target
      newRoute = mapmgr.getPathAwayFrom(startnode, targetWPName, aiPos, player.pos)
      if next(newRoute) == nil then
        targetWPName = nil
      end
    end

    if not newRoute[1] then
      internalState = 'offroad'
      return
    else
      internalState = 'onroad'
    end

    local tempPlan = planAhead(newRoute, currentRoute)
    if tempPlan and (currentRoute == nil or changePlanTimer > 0 or (tempPlan.targetSpeed >= min(aiSpeed, currentRoute.plan.targetSpeed) and targetsCompatible(currentRoute, newRoute))) then
      currentRoute = newRoute
    end
  end

  if currentRoute ~= newRoute then
    planAhead(currentRoute)
  else
    changePlanTimer = max(1, changePlanTimer)
  end
end

local function chasePlan()
  local mapNodes = map.nodes

  local wp1, wp2, dist1 = mapmgr.findClosestRoad(aiPos)
  if wp1 == nil or wp2 == nil then
    internalState = 'offroad'
    return
  end

  if aiDirVec:dot(mapNodes[wp1].pos - mapNodes[wp2].pos) > 0 then wp1, wp2 = wp2, wp1 end

  local plwp1, plwp2, dist2 = mapmgr.findClosestRoad(player.pos)
  if plwp1 == nil or plwp2 == nil then
    internalState = 'offroad'
    return
  end

  if dist1 > max(mapNodes[wp1].radius, mapNodes[wp2].radius) * 2 and dist2 > max(mapNodes[plwp1].radius, mapNodes[plwp2].radius) * 2 then
    internalState = 'offroad'
    return
  end

  if player.dirVec:dot(mapNodes[plwp1].pos - mapNodes[plwp2].pos) > 0 then plwp1, plwp2 = plwp2, plwp1 end

  --chaseData.playerRoad = {plwp1, plwp2}

  local aiPlDist = (aiPos - player.pos):length() -- should this be a signed distance?
  if M.mode == 'follow' then
    setAggressionInternal(min(0.9, 0.3 + 0.002 * aiPlDist))
  else
    setAggressionInternal(min(0.95, 0.7 + 0.001 * aiPlDist))
  end

  -- consider calculating the aggression value but then passing it through a smoother so that transitions between chase mode and follow mode are smooth

  local playerNode = plwp2

  if M.mode == 'follow' and aiSpeed < 0.3 and (wp1 == playerNode or wp2 == playerNode) then -- possibly include oponent velocity?
    currentRoute = nil
  end

  local playerSpeed = player.vel:length()
  if currentRoute then -- route exists
    local curPlan = currentRoute.plan
    local curPath = currentRoute.path
    local playerNodeInPath = waypointInPath(curPath, playerNode, curPlan[2].pathidx)
    local pathLen = getPathLen(currentRoute.path, playerNodeInPath or math.huge) --  curPlan[2].pathidx
    local playerMinPlanLen = getMinPlanLen(0, playerSpeed, staticFrictionCoef * 0.4) -- min(curPlan[#curPlan].speed, curPlan[#curPlan-1].speed, aiSpeed)

    if not playerNodeInPath or (M.mode == 'chase' and pathLen < playerMinPlanLen) then -- think about 'not waypointinPath' condition maybe it needs a different branch
      local newRoute = mapmgr.getPath(wp2, playerNode, driveInLaneFlag and 1e3 or 1) -- maybe wp2 should be curPath[curPlan[2].pathidx]
      --pathLen = getPathLen(newRoute)

      if M.mode == 'chase' then -- and pathLen < playerMinPlanLen
        --local playerSpeed = player.vel:length() -- * 1.1
        if playerSpeed > 1 then -- is this needed?
          --local playerMinPlanLen = getMinPlanLen(0, playerSpeed, staticFrictionCoef * 0.5)
          pathExtend(newRoute, mapmgr.getFleePath(playerNode, player.dirVec, player.pos, playerMinPlanLen, 0, 0)) -- max(minPlanLen-pathLen, playerMinPlanLen) -- mapNodes[plwp1].pos - mapNodes[plwp2].pos
        end
      end

      local tempPlan = planAhead(newRoute, currentRoute)
      if tempPlan and tempPlan.targetSpeed >= min(aiSpeed, currentRoute.plan.targetSpeed) and (tempPlan.targetPos-curPlan.targetPos):dot(aiDirVec) >= 0 then
        currentRoute = newRoute
      else
        planAhead(currentRoute)
      end
    else
      planAhead(currentRoute)
    end

  else -- no current route
    if (M.mode == 'follow' and aiPlDist < 20) or (M.mode == 'chase' and chaseData.suspectState == 'stopped' and (wp1 == playerNode or wp2 == playerNode)) then
      -- do not plan new route if opponent is stopped and ai has reached opponent
      internalState = 'onroad'
      return
    end

    local newRoute = mapmgr.getPath(wp2, playerNode, driveInLaneFlag and 1e3 or 1)

    -- this might be redundant
    if M.mode == 'chase' then -- and pathLen < minPlanLen
      --local pathLen = getPathLen(newRoute)
      --local minPlanLen = getMinPlanLen(0, aiSpeed, staticFrictionCoef * 0.5)
      local minPlanLen = getMinPlanLen(0, playerSpeed, staticFrictionCoef * 0.4)
      pathExtend(newRoute, mapmgr.getFleePath(playerNode, mapNodes[plwp2].pos - mapNodes[plwp1].pos, player.pos, minPlanLen, 0, 0))
    end

    local tempPlan = planAhead(newRoute)
    if tempPlan then
      currentRoute = newRoute
    end
  end

  --chaseData.playerSeg, chaseData.playerXnormOnSeg = posOnPlan(player.pos, currentRoute.plan)

  if M.mode == 'chase' and plPrevVel and wp2 == plwp2 and chaseData.suspectState ~= 'stopped' then
    local playerNodePos1 = mapNodes[plwp2].pos
    local segDir = (playerNodePos1 - mapNodes[plwp1].pos)
    local targetLineDir = vec3(-segDir.y, segDir.x, 0)
    local tarPos = playerNodePos1 + targetLineDir * closestLinePoints(playerNodePos1, playerNodePos1 + targetLineDir, player.pos, player.pos + player.dirVec)

    local p2Target = (tarPos - player.pos):normalized()
    local plVel2Target = player.vel:dot(p2Target)
    local plAccel = (plVel2Target - plPrevVel:dot(p2Target)) / dt
    plAccel = plAccel + sign2(plAccel) * 1e-5
    local plTimeToTarget = (sqrt(max(plVel2Target * plVel2Target + 2 * plAccel * (tarPos - player.pos):length(), 0)) - plVel2Target) / plAccel

    local aiVel2Target = aiVel:dot((tarPos - aiPos):normalized())
    local aiTimeToTarget = (tarPos - aiPos):length() / (aiVel2Target + 1e-30)

    if aiTimeToTarget < plTimeToTarget then
      internalState = 'tail'
      -- return
    else
      internalState = 'onroad'
    end
  else
    internalState = 'onroad'
  end
end

local function warningAIDisabled(message)
  gui.message(message, 5, "AI debug")
  M.mode = 'disabled'
  M.updateGFX = nop
  resetMapAndRoute()
  stateChanged()
end

local function offRoadFollowControl()
  if not player or not player.pos or not aiPos or not aiSpeed then return 0, 0, 0 end

  local ai2PlVec = player.pos - aiPos
  local ai2PlDist = ai2PlVec:length()
  local ai2PlDirVec = ai2PlVec / (ai2PlDist + 1e-30)
  local plSpeedFromAI = player.vel:dot(ai2PlDirVec)
  ai2PlDist = max(0, ai2PlDist - 12)
  local targetSpeed = sqrt(max(0, plSpeedFromAI*plSpeedFromAI*plSpeedFromAI / (abs(plSpeedFromAI) + 1e-30) + 2 * 9.81 * min(1, staticFrictionCoef) * ai2PlDist))
  local speedDif = targetSpeed - aiSpeed
  local throttle = clamp(speedDif, 0, 1)
  local brake = clamp(-speedDif, 0, 1)

  return throttle, brake, targetSpeed
end

M.updateGFX = nop
local function updateGFX(dtGFX)
  dt = dtGFX
  if map == nil then
    if mapmgr.map == nil then return end
    map = mapmgr.map
  end

  -- local cgPos = obj:calcCenterOfGravity()
  -- aiPos:set(cgPos)
  -- aiPos.z = obj:getSurfaceHeightBelow(cgPos)
  local tmpPos = obj:getFrontPosition()
  aiPos:set(tmpPos)
  aiPos.z = max(aiPos.z - 1, obj:getSurfaceHeightBelow(tmpPos))
  prevDirVec:set(aiDirVec)
  aiDirVec:set(obj:getDirectionVector())
  aiUpVec:set(obj:getDirectionVectorUp())
  aiVel:set(obj:getVelocity())
  aiSpeed = aiVel:length()
  aiWidth = aiWidth or obj:getInitialWidth()
  aiLength = aiLength or obj:getInitialLength()
  aiRightVec = aiDirVec:cross(aiUpVec):normalized()
  staticFrictionCoef = 0.95 * obj:getStaticFrictionCoef()

  if trafficBlock.coef > 0 then
    trafficBlock.timer = trafficBlock.timer + dt * trafficBlock.coef
  else
    trafficBlock.timer = trafficBlock.timer * 0.8
  end

  if max(lastCommand.throttle, lastCommand.throttle) > 0.5 and aiSpeed < 1 then
    aiCannotMoveTime = aiCannotMoveTime + dt
  else
    aiCannotMoveTime = 0
  end

  if aiSpeed < 2.5 then
    trafficSide.cTimer = trafficSide.cTimer + dt
    trafficSide.timer = (trafficSide.timer + dt) % (2 * trafficSide.timerRange)
    trafficSide.side = sign2(trafficSide.timerRange - trafficSide.timer)
  else
    trafficSide.cTimer = max(0, trafficSide.cTimer - dt)
    trafficSide.timer = 0
    trafficSide.side = 1
  end

  changePlanTimer = max(0, changePlanTimer - dt)

  ------------------ RANDOM MODE ----------------
  if M.mode == 'random' then
    local newRoute
    if currentRoute == nil or currentRoute.plan[2].pathidx > #currentRoute.path * 0.5 then
      local wp1, wp2 = mapmgr.findClosestRoad(aiPos)
      if wp1 == nil or wp2 == nil then
        warningAIDisabled("Could not find a road network, or closest road is too far")
        return
      end

      if internalState == 'offroad' then
        local vec1 = map.nodes[wp1].pos - aiPos
        local vec2 = map.nodes[wp2].pos - aiPos
        if aiDirVec:dot(vec1) > 0 and aiDirVec:dot(vec2) > 0 then
          if vec1:length() > vec2:length() then wp1, wp2 = wp2, wp1 end
        elseif aiDirVec:dot(map.nodes[wp2].pos - map.nodes[wp1].pos) > 0 then wp1, wp2 = wp2, wp1 end
      elseif aiDirVec:dot(map.nodes[wp2].pos - map.nodes[wp1].pos) > 0 then wp1, wp2 = wp2, wp1 end

      newRoute = mapmgr.getRandomPath(wp1, wp2, driveInLaneFlag and 10e7 or 1)

      if newRoute and newRoute[1] then
        local tempPlan = planAhead(newRoute, currentRoute)
        if tempPlan then
          if not currentRoute then
            currentRoute = newRoute
          else
            local curPlanIdx = currentRoute.plan[2].pathidx
            local curPathCount = #currentRoute.path
            if curPlanIdx >= curPathCount*0.9 or ((tempPlan.targetPos-aiPos):dot(aiDirVec) >= 0 and (curPlanIdx >= curPathCount*0.8 or tempPlan.targetSpeed >= aiSpeed)) then
              currentRoute = newRoute
            end
          end
        end
      end
    end

    if currentRoute ~= newRoute then
      planAhead(currentRoute)
    end

  ------------------ TRAFFIC MODE ----------------
  elseif M.mode == 'traffic' then
    local newRoute
    if currentRoute == nil or aiPos:squaredDistance(map.nodes[currentRoute.path[#currentRoute.path]].pos) < square(getMinPlanLen()) or trafficBlock.timer > trafficBlock.limit then -- getPathLen(currentRoute.path, currentRoute.plan[2].pathidx)
      if currentRoute and currentRoute.path[3] and trafficBlock.timer <= trafficBlock.limit and internalState ~= 'offroad' then
        local path = currentRoute.path
        local pathCount = #path
        local cr0, cr1, cr2 = path[pathCount-2], path[pathCount-1], path[pathCount]
        local cr2Pos = map.nodes[cr2].pos
        local dir1 = (cr2Pos - map.nodes[cr1].pos):normalized()
        local vec = cr2Pos - map.nodes[cr0].pos
        local mirrorOfVecAboutdir1 = (2 * vec:dot(dir1) * dir1 - vec):normalized()
        pathExtend(path, mapmgr.getPathT(cr2, cr2Pos, getMinPlanLen(), 10e4, mirrorOfVecAboutdir1))
      else
        local wp1, wp2 = mapmgr.findClosestRoad(aiPos)

        if wp1 == nil or wp2 == nil then
          warningAIDisabled("Could not find a road network, or closest road is too far")
          return
        end

        local dirVec
        if trafficBlock.timer > trafficBlock.limit and not map.nodes[wp1].links[wp2].oneWay and (map.nodes[wp1].radius + map.nodes[wp2].radius) * 0.5 > 2 then
          dirVec = -aiDirVec
        else
          dirVec = aiDirVec
        end

        if internalState == 'offroad' then
          local vec1 = map.nodes[wp1].pos - aiPos
          local vec2 = map.nodes[wp2].pos - aiPos
          if vec1:squaredLength() > vec2:squaredLength() then
            wp1 = wp2
          end
        else
          wp1 = pickAiWp(wp1, wp2, dirVec)
        end

        -- local newRoute = mapmgr.getRandomPathG(wp1, aiDirVec, getMinPlanLen(), 0.4, 1 / (aiSpeed + 1e-30))
        --newRoute = mapmgr.getRandomPathG(wp1, dirVec, getMinPlanLen(), 0.4, math.huge)
        newRoute = mapmgr.getPathT(wp1, aiPos, getMinPlanLen(), 10e4, aiDirVec)

        if newRoute and newRoute[1] then
          local tempPlan = planAhead(newRoute, currentRoute)

          if tempPlan then
            trafficBlock.limit = math.random() * 10 + 5
            trafficBlock.horn = math.random() >= 0.4 and trafficBlock.limit - (math.random() * 2.5) or math.huge -- horn time start
            trafficSignal.hLimit = math.random() * 2
            intersection.turn = 0

            if not currentRoute or trafficBlock.timer > trafficBlock.limit then
              trafficBlock.timer = 0
              currentRoute = newRoute
            elseif tempPlan.targetSpeed >= aiSpeed and targetsCompatible(currentRoute, newRoute) then
              currentRoute = newRoute
            end
          end
        end
      end
    end

    if currentRoute ~= newRoute then
      planAhead(currentRoute)
      doTrafficActions()
    end

  ------------------ MANUAL MODE ----------------
  elseif M.mode == 'manual' then
    if validateInput(wpList or manualPath) then newManualPath() end

    if aggressionMode == 'rubberBand' then
      updatePlayerData()
      if player ~= nil then
        if (aiPos - player.pos):dot(aiDirVec) > 0 then
          setAggressionInternal(max(min(0.1 + max((150 - (player.pos - aiPos):length())/150, 0), 1), 0.5))
        else
          setAggressionInternal(0.95)
        end
      end
    end

    planAhead(currentRoute)

  ------------------ SPAN MODE ------------------
  elseif M.mode == 'span' then
    if currentRoute == nil then
      local mapNodes = map.nodes
      local wpAft, wpFore = mapmgr.findClosestRoad(aiPos)
      if not (wpAft and wpFore) then
        warningAIDisabled("Could not find a road network, or closest road is too far")
        return
      end
      if aiDirVec:dot(mapNodes[wpFore].pos - mapNodes[wpAft].pos) < 0 then wpAft, wpFore = wpFore, wpAft end

      local target, targetLink

      if not edgeDict then
        -- creates the edgeDict and returns a random edge
        target, targetLink = getMapEdges(M.cutOffDrivability or 0, wpFore)
        if not target then
          warningAIDisabled("No available target with selected characteristics")
          return
        end
      end

      local newRoute = {}

      while true do
        if not target then
          local maxDist = -math.huge
          local lim = 1
          repeat
            -- get most distant non walked edge
            for k, v in pairs(edgeDict) do
              if v <= lim then
                if lim > 1 then edgeDict[k] = 1 end
                local i = string.find(k, '\0')
                local n1id = string.sub(k, 1, i-1)
                local sqDist = (mapNodes[n1id].pos - aiPos):squaredLength()
                if sqDist > maxDist then
                  maxDist = sqDist
                  target = n1id
                  targetLink = string.sub(k, i+1, #k)
                end
              end
            end
            lim = math.huge -- if the first iteration does not produce a target
          until target
        end

        local nodeDegree = 1
        for k, v in pairs(mapNodes[target].links) do
          -- we're looking for neighboring nodes other than the targetLink
          if v ~= targetLink then
            nodeDegree = nodeDegree + 1
          end
        end
        if nodeDegree == 1 then
          local key = target..'\0'..targetLink
          edgeDict[key] = edgeDict[key] + 1
        end

        newRoute = mapmgr.spanMap(wpFore, wpAft, target, edgeDict, driveInLaneFlag and 10e7 or 1)

        if not newRoute[2] and wpFore ~= target then
          -- remove edge from edgeDict list and get a new target (while loop will iterate again)
          edgeDict[target..'\0'..targetLink] = nil
          edgeDict[targetLink..'\0'..target] = nil
          target = nil
          if next(edgeDict) == nil then
            warningAIDisabled("Could not find a path to any of the possible targets")
            return
          end
        elseif not newRoute[1] then
          warningAIDisabled("No Route Found")
          return
        else
          -- insert the second edge node in newRoute if it is not already contained
          local newRouteLen = #newRoute
          if newRoute[newRouteLen-1] ~= targetLink then newRoute[newRouteLen+1] = targetLink end
          break
        end
      end

      if planAhead(newRoute) == nil then return end
      currentRoute = newRoute
    else
      planAhead(currentRoute)
    end

  ------------------ FLEE MODE ------------------
  elseif M.mode == 'flee' then
    updatePlayerData()
    if player then
      if validateInput() then
        targetWPName = wpList[1]
        wpList = nil
      end

      setAggressionInternal(min(1, max(0.3, 1 - 0.002 * (player.pos - aiPos):length())))

      fleePlan()

      if internalState == 'offroad' then
        local targetPos = aiPos + (aiPos - player.pos) * 100
        local targetSpeed = math.huge
        driveToTarget(targetPos, 1, 0, targetSpeed)
        return
      end
    else
      --gui.message("No vehicle to Flee from", 5, "AI debug") -- TODO: this freezes the up because it runs on the gfx step
      return
    end

  ------------------ CHASE MODE ------------------
  elseif M.mode == 'chase' or M.mode == 'follow' then
    updatePlayerData()
    if player then

      if player.vel:squaredLength() < 2.25 then
        chaseData.suspectStoppedTimer = chaseData.suspectStoppedTimer + dt
      else
        chaseData.suspectStoppedTimer = 0
      end

      if chaseData.suspectStoppedTimer > 10 then
        chaseData.suspectState = 'stopped'
      else
        chaseData.suspectState = nil
      end

      chasePlan()

      if plPrevVel then
        plPrevVel:set(player.vel)
      else
        plPrevVel = vec3(player.vel)
      end

      if internalState == 'tail' then
        --internalState = 'onroad'
        --currentRoute = nil
        local plai = player.pos - aiPos
        local relvel = aiVel:dot(plai) - player.vel:dot(plai)
        if relvel > 0 then
          driveToTarget(player.pos + (plai:squaredLength() / (relvel + 1e-30)) * player.vel, 1, 0, math.huge)
        else
          driveToTarget(player.pos, 1, 0, math.huge)
        end
        return
      elseif internalState == 'offroad' then
        if M.mode == 'follow' then
          local throttle, brake, targetSpeed = offRoadFollowControl()
          driveToTarget(player.pos, throttle, brake, targetSpeed)
        else
          driveToTarget(player.pos, 1, 0, math.huge)
        end
        return
      elseif currentRoute == nil then
        driveCar(0, 0, 0, 1)
        return
      end

    else
      --gui.message("No vehicle to Chase", 5, "AI debug")
      return
    end

  ------------------ STOP MODE ------------------
  elseif M.mode == 'stop' then
    if aiVel:dot(aiDirVec) > 0 then
      driveCar(0, 0, 0.5, 0)
    else
      driveCar(0, 1, 0, 0)
    end
    if aiSpeed < 0.08 then --  or aiVel:dot(aiDirVec) < 0
      driveCar(0, 0, 0, 0) -- only parkingbrake
      M.mode = 'disabled'
      M.manualTargetName = nil
      M.updateGFX = nop
      resetMapAndRoute()
      stateChanged()
      return
    end
  end
  -----------------------------------------------

  if currentRoute then
    local plan = currentRoute.plan
    local targetPos = plan.targetPos
    local aiSeg = plan.aiSeg

    -- cleanup path if it has gotten too long
    if not race and plan[aiSeg].pathidx >= 10 and currentRoute.path[20] then
      local newPath = {}
      local j, k = 0, plan[aiSeg].pathidx
      for i = k, #currentRoute.path do
        j = j + 1
        newPath[j] = currentRoute.path[i]
      end
      currentRoute.path = newPath

      k = k - 1
      for i = 1, #plan do
        plan[i].pathidx = plan[i].pathidx - k
      end
    end

    local targetSpeed = plan.targetSpeed
    trafficBlock.coef = plan.trafficBlockCoef or trafficBlock.coef

    if aiUpVec:dot(gravityVec) > 0 then -- vehicle upside down
      return
    end

    if not targetPos or ((targetSpeed < 0.05 and aiSpeed < 0.1 or (targetPos - aiPos):dot(aiDirVec) < 0) and aiSeg == #plan-1) then
      if M.mode == 'span' then
        local path = currentRoute.path
        for i = 1, #path - 1 do
          local key = path[i]..'\0'..path[i+1]
          -- in case we have gone over an edge that is not in the edgeDict list
          edgeDict[key] = edgeDict[key] and (edgeDict[key] * 20)
        end
      end

      driveCar(0, 0, 0, 1)
      aistatus('route done', 'route')
      gui.message("Route done", 5, "AI debug")
      currentRoute = nil
      speedRecordings = {}
      return
    end

      -- TODO: this still runs if there is no currentPlan, but raises error if there is no targetSpeed
    if not controller.isFrozen and aiSpeed < 0.1 and targetSpeed > 0.5 and (lastCommand.throttle ~= 0 or lastCommand.brake ~= 0) then
      crashTime = crashTime + dt
      if crashTime > 1 then
        crashDir = vec3(aiDirVec)
        crashManeuver = 1
      end
    else
      crashTime = 0
    end

    -- Throttle and Brake control
    local speedDif = targetSpeedSmoother:get(targetSpeed, dt) - aiSpeed
    local lowSpeedDif = (speedDif - clamp((aiSpeed - 2) * 0.5, 0, 1)) * 0.5
    local throttle = clamp(lowSpeedDif, 0, 1)
    local brake = clamp(-speedDif, 0, 1)

    driveToTarget(targetPos, throttle, brake)
  end
end


local function debugDraw(x, y, z)
  local debugDrawer = obj.debugDrawProxy
  local focusPos = float3(x, y, z)

  if M.mode == 'script' and scriptai ~= nil then
    scriptai.debugDraw()
  end

  if currentRoute then
    local plan = currentRoute.plan
    local targetPos = plan.targetPos
    local targetSpeed = plan.targetSpeed
    if targetPos then
      --debugDrawer:drawSphere(0.25, (aiPos - aiDirVec*aiLength):toFloat3(), color(0,255,0,255))
      --local aiXnormOnSeg = max(0, aiPos:xnormOnLine(plan[aiSeg].pos, plan[aiSeg+1].pos))
      local aiXnormOnSeg = plan.aiXnormOnSeg
      local aiSeg = plan.aiSeg
      debugDrawer:drawSphere(0.25, targetPos:toFloat3(), color(255,0,0,255))
      local shadowPos = currentRoute.plan[aiSeg].pos + aiXnormOnSeg*(plan[aiSeg+1].pos - plan[aiSeg].pos)
      debugDrawer:drawSphere(0.25, shadowPos:toFloat3(), color(0,0,255,255))

      for plID, v in pairs(mapmgr.objects) do
        if plID ~= objectId then
          local plPosFront = vec3(obj:getObjectFrontPosition(plID))
          debugDrawer:drawSphere(0.25, plPosFront:toFloat3(), color(0,0,255,255))
        end
      end

      if player then
        debugDrawer:drawSphere(0.3, player.pos:toFloat3(), color(0,255,0,255))
      end
    end

    if M.debugMode == 'target' then
      if map ~= nil and map.nodes and currentRoute.path then
        local p = map.nodes[currentRoute.path[#currentRoute.path]].pos:toFloat3()
        --debugDrawer:drawSphere(4, p, color(255,0,0,100))
        --debugDrawer:drawText(p + float3(0, 0, 4), color(0,0,0,255), 'Destination')
      end

    elseif M.debugMode == 'route' then
      if currentRoute.path and next(map) ~= nil then
        local p = map.nodes[currentRoute.path[#currentRoute.path]].pos:toFloat3()
        debugDrawer:drawSphere(4, p, color(255,0,0,100))
        debugDrawer:drawText(p + float3(0, 0, 4), color(0,0,0,255), 'Destination')
      end

      local maxLen = 700
      local last = routeRec.last
      local len = min(#routeRec, maxLen)
      if len == 0 or (routeRec[last] - aiPos:toFloat3()):length() > 7 then
        last = 1 + last % maxLen
        routeRec[last] = aiPos:toFloat3()
        len = min(len+1, maxLen)
        routeRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      local col = color(0,0,0,128)
      for i = 1, len-1 do
        debugDrawer:drawSquarePrism(routeRec[1+(last+i-1)%len], routeRec[1+(last+i)%len], fl3, fl3, col)
      end

      if currentRoute.plan[1].pathidx then
        local mapNodes = map.nodes
        local path = currentRoute.path
        fl3 = fl3 + float3(0, aiWidth, 0)
        local col = color(255,0,0,120)
        for i = currentRoute.plan[1].pathidx, #path - 1 do
          debugDrawer:drawSquarePrism(mapNodes[path[i]].pos:toFloat3(), mapNodes[path[i+1]].pos:toFloat3(), fl3, fl3, col)
        end
      end

    elseif M.debugMode == 'speeds' then
      if plan[1] then
        local red = color(255,0,0,200) -- getContrastColor(objectId)
        local prevPoint = plan[1].pos:toFloat3()
        local prevSpeed = -1
        local drawLen = 0
        for i = 1, #plan do
          local n = plan[i]
          local p = n.pos:toFloat3()
          local v = (n.speed >= 0 and n.speed) or prevSpeed
          local p1 = p + float3(0, 0, v*0.2)
          --debugDrawer:drawLine(p + float3(0, 0, v*0.2), (n.pos + n.turnDir):toFloat3() + float3(0, 0, v*0.2), col)
          debugDrawer:drawCylinder(p, p1, 0.03, red)
          local col = red
          --if chaseData.playerSeg and chaseData.playerSeg + 1 == i then col = color(0,0,0,200) end
          debugDrawer:drawCylinder(prevPoint, p1, 0.05, col)
          debugDrawer:drawText(p1, color(0,0,0,255), strFormat("%2.0f", v*3.6) .. " kph")
          --debugDrawer:drawSphere(n.radius, p, color(128, 128, 128, 128))
          prevPoint, prevSpeed = p1, v
          --drawLen = drawLen + n.vec:length()
          -- if traffic and traffic[i] then
          --   for _, data in ipairs(traffic[i]) do
          --     local plPosOnPlan = linePointFromXnorm(n.pos, plan[i+1].pos, data[2])
          --     debugDrawer:drawSphere(0.25, plPosOnPlan:toFloat3(), color(0,255,0,100))
          --   end
          -- end
          -- if drawLen > 150 then break end
        end

        local aiPosFlt = aiPos:toFloat3()
        local speedRecLen = #speedRecordings

        if speedRecLen == 0 or (speedRecordings[speedRecLen][1]-aiPosFlt):length() > 0.25 then
          tableInsert(speedRecordings, {aiPosFlt, aiSpeed, targetSpeed, lastCommand.brake, lastCommand.throttle})
          speedRecLen = speedRecLen + 1
        end

        local lastEntry
        local zOffSet = float3(0,0,0.5)
        local yellow, blue = color(255,255,0,200), color(0,0,255,200)
        for i = 1, speedRecLen do
          local v = speedRecordings[i]
          if lastEntry then
            -- actuall speed
            debugDrawer:drawCylinder(lastEntry[1]+float3(0,0,lastEntry[2]*0.2), v[1]+float3(0, 0, v[2]*0.2), 0.02, yellow)
            -- target speed
            debugDrawer:drawCylinder(lastEntry[1]+float3(0,0,lastEntry[3]*0.2), v[1]+float3(0, 0, v[3]*0.2), 0.02, blue)
          end
          debugDrawer:drawCylinder(v[1], v[1]+float3(0, 0, v[3]*0.2), 0.01, blue)

          if (focusPos - v[1]):length() < labelRenderDistance then
            debugDrawer:drawText(v[1]+float3(0, 0, v[2]*0.2)+zOffSet, yellow, strFormat("%2.0f", v[2]*3.6) .. " kph")
            debugDrawer:drawText(v[1]+float3(0, 0, v[3]*0.2)+zOffSet, blue, strFormat("%2.0f", v[3]*3.6) .. " kph")
          end
          lastEntry = v
        end

        if speedRecLen > 175 then
          tableRemove(speedRecordings, 1)
        end
      end

      -- Debug Throttle brake application
      local maxLen = 175
      local len = min(#trajecRec, maxLen)
      local last = trajecRec.last
      local aiPosFlt = aiPos:toFloat3()
      if len == 0 or (trajecRec[last][1]-aiPosFlt):length() > 0.25 then
        last = 1 + last % maxLen
        trajecRec[last] = {aiPosFlt, lastCommand.throttle, lastCommand.brake}
        len = min(len+1, maxLen)
        trajecRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      for i = 1, len-1 do
        local n = trajecRec[1+(last+i)%len]
        debugDrawer:drawSquarePrism(trajecRec[1+(last+i-1)%len][1], n[1], fl3, fl3, color(255*sqrt(abs(n[3])), 255*sqrt(n[2]), 0, 100))
      end

      -- Player segment visual debug for chase / follow mode
      -- if chaseData.playerRoad then
      --   local col1, col2
      --   if internalState == 'tail' then
      --     col1 = color(0,0,0,200)
      --     col2 = color(0,0,0,200)
      --   else
      --     col1 = color(255,0,0,100)
      --     col2 = color(0,0,255,100)
      --   end
      --   local plwp1 = chaseData.playerRoad[1]
      --   debugDrawer:drawSphere(2, map.nodes[plwp1].pos:toFloat3(), col1)
      --   local plwp2 = chaseData.playerRoad[2]
      --   debugDrawer:drawSphere(2, map.nodes[plwp2].pos:toFloat3(), col2)
      -- end

    elseif M.debugMode == 'trajectory' then
      -- Debug Curvatures
      -- local plan = currentRoute.plan
      -- if plan ~= nil then
      --   local prevPoint = plan[1].pos:toFloat3()
      --   for i = 1, #plan do
      --     local p = plan[i].pos:toFloat3()
      --     local v = plan[i].curvature or 1e-10
      --     local scaledV = abs(1000 * v)
      --     debugDrawer:drawCylinder(p, p + float3(0, 0, scaledV), 0.06, color(abs(min(fsign(v),0))*255,max(fsign(v),0)*255,0,200))
      --     debugDrawer:drawText(p + float3(0, 0, scaledV), color(0,0,0,255), strFormat("%5.4e", v))
      --     debugDrawer:drawCylinder(prevPoint, p + float3(0, 0, scaledV), 0.06, col)
      --     prevPoint = p + float3(0, 0, scaledV)
      --   end
      -- end

      -- Debug Planned Speeds
      if plan[1] then
        local col = getContrastColor(objectId)
        local prevPoint = plan[1].pos:toFloat3()
        local prevSpeed = -1
        local drawLen = 0
        for i = 1, #plan do
          local n = plan[i]
          local p = n.pos:toFloat3()
          local v = (n.speed >= 0 and n.speed) or prevSpeed
          local p1 = p + float3(0, 0, v*0.2)
          --debugDrawer:drawLine(p + float3(0, 0, v*0.2), (n.pos + n.turnDir):toFloat3() + float3(0, 0, v*0.2), col)
          debugDrawer:drawCylinder(p, p1, 0.03, col)
          debugDrawer:drawCylinder(prevPoint, p1, 0.05, col)
          debugDrawer:drawText(p1, color(0,0,0,255), strFormat("%2.0f", v*3.6) .. " kph")
          prevPoint = p1
          prevSpeed = v
          drawLen = drawLen + n.vec:length()
          if drawLen > 80 then break end
        end
      end

      -- Debug Throttle brake application
      local maxLen = 175
      local len = min(#trajecRec, maxLen)
      local last = trajecRec.last
      if len == 0 or (trajecRec[last][1] - aiPos:toFloat3()):length() > 0.25 then
        last = 1 + last % maxLen
        trajecRec[last] = {aiPos:toFloat3(), lastCommand.throttle, lastCommand.brake}
        len = min(len+1, maxLen)
        trajecRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      for i = 1, len-1 do
        local n = trajecRec[1+(last+i)%len]
        debugDrawer:drawSquarePrism(trajecRec[1+(last+i-1)%len][1], n[1], fl3, fl3, color(255*sqrt(abs(n[3])), 255*sqrt(n[2]), 0, 100))
      end
    end
  end
end

local function init()
  map = {}
end

local function setAvoidCars(v)
  M.avoidCarsMaster = v
  if M.avoidCarsMaster == 'off' or M.avoidCarsMaster == 'on' then
    avoidCars = M.avoidCarsMaster
  else
    avoidCars = M.mode == 'manual' and 'off' or 'on'
  end
  stateChanged()
end

local function driveInLane(v)
  if v == 'on' then
    M.driveInLaneFlag = 'on'
    driveInLaneFlag = true
  else
    M.driveInLaneFlag = 'off'
    driveInLaneFlag = false
  end
  stateChanged()
end

local function setMode(mode)
  local previousMode = M.mode

  if M.avoidCarsMaster == 'off' or M.avoidCarsMaster == 'on' then
    avoidCars = M.avoidCarsMaster
  else
    avoidCars = (mode == 'manual' or (mode == nil and M.mode == 'manual')) and 'off' or 'on'
  end

  if mode ~= nil then
    M.mode = mode
  end

  if M.mode ~= 'script' then
    if M.mode ~= 'disabled' then -- TODO: this should be more explicit maybe?
      resetMapAndRoute()

      mapmgr.requestMap()
      M.updateGFX = updateGFX
    end

    if M.mode == 'disabled' then
      if previousMode ~= M.mode then
        driveCar(0, 0, 0, 0)
      end
      M.updateGFX = nop
      currentRoute = nil
    elseif M.mode ~= 'stop' then
      if controller.mainController then
        controller.mainController.setGearboxMode("arcade")
      end
    end

    if M.mode == 'traffic' then
      setSpeedMode('legal')
      obj:setSelfCollisionMode(2)
      obj:setAerodynamicsMode(2)
    else
      obj:setSelfCollisionMode(1)
      obj:setAerodynamicsMode(1)
    end

    stateChanged()
  end

  speedRecordings = {}
  trajecRec = {last = 0}
  routeRec = {last = 0}
end

local function reset() -- called when the user pressed I
  M.manualTargetName = nil
  trafficBlock = {timer = 0, coef = 0, limit = 6, horn = 0}

  trafficSide.timer = 0
  trafficSide.cTimer = 0
  trafficSide.side = 1

  trafficSignal = {hTimer = 0, hLimit = 1}
  intersection = {timer = 0, turn = 0}

  crashManeuver = 0
  threewayturn = {state = 0, speedDifInt = 0}
  if M.mode ~= 'disabled' then
    driveCar(0, 0, 0, 0)
  end
  setMode() -- some scenarios don't work if this is changed to setMode('disabled')
  stateChanged()
end

local function resetLearning()
end

local function setVehicleDebugMode(newMode)
  tableMerge(M, newMode)
  if M.debugMode ~= 'trajectory' then
    trajecRec = {last = 0}
  end
  if M.debugMode ~= 'route' then
    routeRec = {last = 0}
  end
  if M.debugMode ~= 'speeds' then
    speedRecordings = {}
  end
  if M.debugMode ~= 'off' then
    M.debugDraw = debugDraw
  else
    M.debugDraw = nop
  end
end

local function setState(newState)
  tableMerge(M, newState)
  setAggressionExternal(M.extAggression)
  setMode()
  setVehicleDebugMode(M)
  setTargetObjectID(M.targetObjectID)
end

local function setTarget(wp)
  M.manualTargetName = wp
  validateInput = validateUserInput
  wpList = {wp}
end

local function setPath(path)
  manualPath = path
  validateInput = validateUserInput
end

local function driveUsingPath(arg)
  --[[ At least one argument of either path or wpTargetList must be specified. All other arguments are optional.

  * path: A sequence of waypoint names that form a path by themselves to be followed in the order provided.
  * wpTargetList: Type: A sequence of waypoint names to be used as succesive targets ex. wpTargetList = {'wp1', 'wp2'}.
                  Between any two consequitive waypoints a shortest path route will be followed.

  -- Optional Arguments --
  * wpSpeeds: Type: (key/value pairs, key: "node_name", value: speed, number in m/s)
              Define target speeds for individual waypoints. The ai will try to meet this speed when at the given waypoint.
  * noOfLaps: Type: number. Default value: nil
              The number of laps if the path is a loop. If not defined, the ai will just follow the succesion of waypoints once.
  * routeSpeed: A speed in m/s. To be used in tandem with "routeSpeedMode".
                Type: number
  * routeSpeedMode: Values: 'limit': the ai will not go above the 'routeSpeed' defined by routeSpeed.
                            'set': the ai will try to always go at the speed defined by "routeSpeed".
  * driveInLane: Values: 'on' (anything else is considered off/inactive)
                 When 'on' the ai will keep on the correct side of the road on two way roads.
                 This also affects pathFinding in that when this option is active ai paths will traverse roads in the legal direction if posibble.
                 Default: inactive
  * aggression: Value: 0.3 - 1. The aggression value with which the ai will drive the route.
                At 1 the ai will drive at the limit of traction. A value of 0.3 would be considered normal every day driving, going shopping etc.
                Default: 0.3
  * avoidCars: Values: 'on' / 'off'.  When 'on' the ai will be aware of (avoid crashing into) other vehicles on the map. Default is 'off'
  * examples:
  ai.driveUsingPath{ wpTargetList = {'wp1', 'wp10'}, driveInLane = 'on', avoidCars = 'on', routeSpeed = 35, routeSpeedMode = 'limit', wpSpeeds = {wp1 = 10, wp2 = 40}, aggression = 0.3}
  In the above example the speeds set for wp1 and wp2 will take precedence over "routeSpeed" for the specified nodes.
  --]]

  if (arg.wpTargetList == nil and arg.path == nil and arg.script == nil) or
    (type(arg.wpTargetList) ~= 'table' and type(arg.path) ~= 'table' and type(arg.script) ~= 'table') or
    (arg.wpSpeeds ~= nil and type(arg.wpSpeeds) ~= 'table') or
    (arg.noOfLaps ~= nil and type(arg.noOfLaps) ~= 'number') or
    (arg.routeSpeed ~= nil and type(arg.routeSpeed) ~= 'number') or
    (arg.routeSpeedMode ~= nil and type(arg.routeSpeedMode) ~= 'string') or
    (arg.driveInLane ~= nil and type(arg.driveInLane) ~= 'string') or
    (arg.aggression ~= nil and type(arg.aggression) ~= 'number')
  then
    return
  end

  if arg.resetLearning then
    resetLearning()
  end

  setState({mode = 'manual'})

  noOfLaps = arg.noOfLaps and max(arg.noOfLaps, 1) or 1
  wpList = arg.wpTargetList
  manualPath = arg.path
  validateInput = validateUserInput
  avoidCars = arg.avoidCars or 'off'

  if noOfLaps > 1 and wpList[2] and wpList[1] == wpList[#wpList] then
    race = true
  end

  speedList = arg.wpSpeeds or {}
  setSpeed(arg.routeSpeed)
  setSpeedMode(arg.routeSpeedMode)

  driveInLane(arg.driveInLane)

  setAggressionExternal(arg.aggression)
  stateChanged()
end

local function spanMap(cutOffDrivability)
  M.cutOffDrivability = cutOffDrivability or 0
  setState({mode = 'span'})
  stateChanged()
end

local function setCutOffDrivability(drivability)
  M.cutOffDrivability = drivability or 0
  stateChanged()
end

local function onDeserialized(v)
  setState(v)
  stateChanged()
end

local function dumpCurrentRoute()
  dump(currentRoute)
end

local function startRecording()
  M.mode = 'script'
  scriptai = require("scriptai")
  scriptai.startRecording()
  M.updateGFX = scriptai.updateGFX
end

local function stopRecording()
  M.mode = 'disabled'
  scriptai = require("scriptai")
  local script = scriptai.stopRecording()
  M.updateGFX = scriptai.updateGFX
  return script
end

local function startFollowing(...)
  M.mode = 'script'
  scriptai = require("scriptai")
  scriptai.startFollowing(...)
  M.updateGFX = scriptai.updateGFX
end

local function scriptStop(...)
  M.mode = 'disabled'
  scriptai = require("scriptai")
  scriptai.scriptStop(...)
  M.updateGFX = scriptai.updateGFX
end

local function scriptState()
  scriptai = require("scriptai")
  return scriptai.scriptState()
end

local function setScriptDebugMode(mode)
  scriptai = require("scriptai")
  if mode == nil or mode == 'off' then
    M.debugMode = 'all'
    M.debugDraw = nop
    return
  end

  M.debugDraw = debugDraw
  scriptai.debugMode = mode
end

local function isDriving()
  return M.updateGFX == updateGFX or (scriptai ~= nil and scriptai.isDriving())
end

-- public interface
M.driveInLane = driveInLane
M.stateChanged = stateChanged
M.reset = reset
M.init = init
M.setMode = setMode
M.setAvoidCars = setAvoidCars
M.setTarget = setTarget
M.setPath = setPath
M.setSpeed = setSpeed
M.setSpeedMode = setSpeedMode
M.setVehicleDebugMode = setVehicleDebugMode
M.setState = setState
M.getState = getState
M.debugDraw = nop
M.driveUsingPath = driveUsingPath
M.setAggressionMode = setAggressionMode
M.setAggression = setAggressionExternal
M.setForceLateral = setForceLateral
M.setDistPastEdge = setDistPastEdge
M.onDeserialized = onDeserialized
M.setTargetObjectID = setTargetObjectID
M.dumpCurrentRoute = dumpCurrentRoute
M.spanMap = spanMap
M.setCutOffDrivability = setCutOffDrivability
M.resetLearning = resetLearning
M.isDriving = isDriving

-- scriptai
M.startRecording = startRecording
M.stopRecording = stopRecording
M.startFollowing = startFollowing
M.stopFollowing = scriptStop
M.scriptStop = scriptStop
M.scriptState = scriptState
M.setScriptDebugMode = setScriptDebugMode
return M
